/*
	This file is part of CrabEmu.

	Copyright (C) 2005 Lawrence Sebald

	CrabEmu is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	CrabEmu is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CrabEmu; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include "sn76489.h"

static const float CLOCKS_PER_SAMPLE = (3579545.0f / 16.0f / 44100.0f);

/* These constants came from Maxim's core. */
static const int volume_values[16] =
	{ 892, 774, 669, 575, 492, 417, 351, 292,
	  239, 192, 150, 113,  80,  50,  24,   0 };

int sn76489_init(sn76489_t *psg)	{
	psg->volume[0] = 0xF;
	psg->volume[1] = 0xF;
	psg->volume[2] = 0xF;
	psg->volume[3] = 0xF;

	psg->tone[0] = 0x00;
	psg->tone[1] = 0x00;
	psg->tone[2] = 0x00;
	psg->noise = 0x0;

	psg->latched_reg = LATCH_TONE0;

	psg->counter[0] = 0x00;
	psg->counter[1] = 0x00;
	psg->counter[2] = 0x00;
	psg->counter[3] = 0x00;

	psg->tone_state[0] = 1;
	psg->tone_state[1] = 1;
	psg->tone_state[2] = 1;
	psg->tone_state[3] = 1;

	psg->enabled_channels = 0x0F;

	return 0;
}

void sn76489_write(sn76489_t *psg, unsigned char byte)	{
	if(byte & 0x80)		{
		/* This is a LATCH/DATA byte */
		psg->latched_reg = (byte & 0x70);

		switch(psg->latched_reg)	{
			case LATCH_TONE0:
				psg->tone[0] = (psg->tone[0] & 0x3F0) | (byte & 0x0F);
				break;
			case LATCH_TONE1:
				psg->tone[1] = (psg->tone[1] & 0x3F0) | (byte & 0x0F);
				break;
			case LATCH_TONE2:
				psg->tone[2] = (psg->tone[2] & 0x3F0) | (byte & 0x0F);
				break;
			case LATCH_NOISE:
				psg->noise = (byte & 0x0F);
				break;
			case LATCH_VOL0:
				psg->volume[0] = (byte & 0x0F);
				break;
			case LATCH_VOL1:
				psg->volume[1] = (byte & 0x0F);
				break;
			case LATCH_VOL2:
				psg->volume[2] = (byte & 0x0F);
				break;
			case LATCH_VOL3:
				psg->volume[3] = (byte & 0x0F);
				break;
		}
	}
	else	{
		/* This is a DATA byte */
		switch(psg->latched_reg)	{
			case LATCH_TONE0:
				psg->tone[0] = (psg->tone[0] & 0x000F) | ((byte & 0x3F) << 4);
				break;
			case LATCH_TONE1:
				psg->tone[1] = (psg->tone[1] & 0x000F) | ((byte & 0x3F) << 4);
				break;
			case LATCH_TONE2:
				psg->tone[2] = (psg->tone[2] & 0x000F) | ((byte & 0x3F) << 4);
				break;
			case LATCH_NOISE:
				psg->noise = (byte & 0x0F);
				break;
			case LATCH_VOL0:
				psg->volume[0] = (byte & 0x0F);
				break;
			case LATCH_VOL1:
				psg->volume[1] = (byte & 0x0F);
				break;
			case LATCH_VOL2:
				psg->volume[2] = (byte & 0x0F);
				break;
			case LATCH_VOL3:
				psg->volume[3] = (byte & 0x0F);
				break;
		}
	}
}

/* Assume 44100 hz mono sound for now, this can be adjusted later (but it makes
   calculations easier for now. */
unsigned long sn76489_execute(sn76489_t *psg, signed short *buf, unsigned long cpu_cycles)	{
	unsigned long samples = cpu_cycles / 16 / (unsigned long) CLOCKS_PER_SAMPLE; /* How many samples will we calculate */
	signed int channels[4];
	unsigned long i, j;
	
	for(i = 0; i < samples; ++i)	{
		for(j = 0; j < 3; ++j)	{
			psg->counter[j] -= CLOCKS_PER_SAMPLE;
			channels[j] = ((psg->enabled_channels & (1 << j)) >> j) * psg->tone_state[j]  * volume_values[psg->volume[j]];

			if(psg->counter[j] <= 0.0f)	{
				psg->tone_state[j] = -psg->tone_state[j];
				psg->counter[j] += psg->tone[j];
			}
		}

		buf[i] = (channels[0] + channels[1] + channels[2]) & 0xFFFF;
	}

	return samples;
}

void sn76489_execute_samples(sn76489_t *psg, signed short *buf, unsigned long samples)	{
	signed int channels[4];
	unsigned long i, j;

	for(i = 0; i < samples; ++i)	{
		for(j = 0; j < 3; ++j)	{
			psg->counter[j] -= CLOCKS_PER_SAMPLE;
				channels[j] = ((psg->enabled_channels & (1 << j)) >> j) * psg->tone_state[j]  * volume_values[psg->volume[j]];
			if(psg->counter[j] <= 0.0f)	{
				psg->tone_state[j] = -psg->tone_state[j];
				psg->counter[j] += psg->tone[j];
			}
		}
		
		buf[i] = (channels[0] + channels[1] + channels[2]) & 0xFFFF;
	}
}
