/*
	This file is part of CrabEmu.

	Copyright (C) 2005 Lawrence Sebald

	CrabEmu is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	CrabEmu is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CrabEmu; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "z80.h"
#include "tables.h"
#include <stdio.h>
#include <assert.h>

static uint8 z80_dummy_read(uint16 addr)	{
	return 0;
}

static void z80_dummy_write(uint16 addr, uint8 data)	{
}

void z80_set_portread(z80_t *cpu, uint8 (*pread)(uint16 port))	{
	if(pread == NULL)
		cpu->pread = z80_dummy_read;
	else
		cpu->pread = pread;
}

void z80_set_memread(z80_t *cpu, uint8 (*mread)(uint16 addr))	{
	if(mread == NULL)
		cpu->mread = z80_dummy_read;
	else
		cpu->mread = mread;
}

void z80_set_portwrite(z80_t *cpu, void (*pwrite)(uint16 port, uint8 data))	{
	if(pwrite == NULL)
		cpu->pwrite = z80_dummy_write;
	else
		cpu->pwrite = pwrite;
}

void z80_set_memwrite(z80_t *cpu, void (*mwrite)(uint16 addr, uint8 data))	{
	if(mwrite == NULL)
		cpu->mwrite = z80_dummy_write;
	else
		cpu->mwrite = mwrite;
}

void z80_init(z80_t *cpu)	{
	cpu->pread = z80_dummy_read;
	cpu->mread = z80_dummy_read;
	cpu->pwrite = z80_dummy_write;
	cpu->mwrite = z80_dummy_write;
}

void z80_reset(z80_t *cpu)	{
	cpu->pc.w = 0x0000;
	cpu->iff1 = 0;
	cpu->iff2 = 0;
	cpu->im = 0;
	cpu->halt = 0;
	cpu->r_top = 0;
	cpu->af.w = 0x0040;
	cpu->bc.w = 0x0000;
	cpu->de.w = 0x0000;
	cpu->hl.w = 0x0000;
	cpu->ix.w = 0xFFFF;
	cpu->iy.w = 0xFFFF;
	cpu->sp.w = 0x0000;
	cpu->ir.w = 0x0000;
	cpu->afp.w = 0x0000;
	cpu->bcp.w = 0x0000;
	cpu->dep.w = 0x0000;
	cpu->hlp.w = 0x0000;
	cpu->cycles = 0;
}

void z80_pulse_nmi(z80_t *cpu)	{
	if(cpu->halt)	{
		cpu->pc.w++;
		cpu->halt = 0;
	}

	cpu->iff2 = cpu->iff1;
	cpu->iff1 = 0;
	cpu->mwrite(--cpu->sp.w, cpu->pc.b.h);
	cpu->mwrite(--cpu->sp.w, cpu->pc.b.l);
	cpu->pc.w = 0x0066;
	cpu->cycles += 11;

	cpu->ir.b.l = (cpu->ir.b.l & 0x80) | ((cpu->ir.b.l + 1) & 0x7F);
}

void z80_pulse_irq(z80_t *cpu)	{
	if(cpu->iff1 == 0) /* interrupts are disabled. */
		return;

	if(cpu->halt)	{
		cpu->pc.w++;
		cpu->halt = 0;
	}

	cpu->iff1 = cpu->iff2 = 0;

	switch(cpu->im)	{
		case 1:
			cpu->mwrite(--cpu->sp.w, cpu->pc.b.h);
			cpu->mwrite(--cpu->sp.w, cpu->pc.b.l);
			cpu->pc.w = 0x0038;
			cpu->cycles += 13;
			break;
		default:
			printf("Unimplemented Interrupt Mode: %d\n", cpu->im);
			assert(0);
	}
}

uint32 z80_execute(z80_t *cpu, uint32 cycles)	{
	uint8 inst;
	uint32 real_cycles;

	while(cpu->cycles < cycles)	{
		cpu->ei = 0;
		inst = cpu->mread(cpu->pc.w++);
		++cpu->ir.b.l;

#		include "z80_switch_of_doom.h"

	}

	real_cycles = cpu->cycles;
	cpu->cycles = 0;

	return real_cycles;
}
