/*
	This file is part of CrabEmu.

	Copyright (C) 2005 Lawrence Sebald

	CrabEmu is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	CrabEmu is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CrabEmu; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <AudioUnit/AudioUnit.h>
#include <pthread.h>
#include "sound.h"

static AudioUnit outputAU;
static unsigned char buffer[44100 * 2];
static int buffer_size = 0;
static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t cond = PTHREAD_COND_INITIALIZER;

static OSStatus sound_callback(void *inRefCon, 
                               AudioUnitRenderActionFlags inActionFlags,
                               const AudioTimeStamp *inTimeStamp,
                               UInt32 inBusNumber,
                               AudioBuffer *ioData)	{
	UInt32 len = ioData->mDataByteSize;
	void *ptr = ioData->mData;

	if(buffer_size == 0)	{
		memset(ioData->mData, 0, len);
		return noErr;
	}

	pthread_mutex_lock(&mutex);

	if(buffer_size < 1024)
		pthread_cond_wait(&cond, &mutex);

	memcpy(ptr, buffer, len);
	buffer_size -= len;

	if(buffer_size)
		memcpy(buffer, buffer + buffer_size, buffer_size);

	pthread_mutex_unlock(&mutex);

	return noErr;
}

void sound_update_buffer(signed short *buf, int length)	{
	pthread_mutex_lock(&mutex);

	memcpy(buffer + buffer_size, buf, length);
	buffer_size += length;

	if(buffer_size >= 1024)	{
		pthread_cond_signal(&cond);
	}

	pthread_mutex_unlock(&mutex);
}

int sound_init()	{
	OSStatus error = noErr;
	ComponentDescription desc;
	AudioStreamBasicDescription basic_desc;
	Component comp;
	struct AudioUnitInputCallback callback;

	/* Set up the AudioStreamBasicDescription */
	basic_desc.mFormatID = kAudioFormatLinearPCM;
	basic_desc.mFormatFlags = kLinearPCMFormatFlagIsPacked | kLinearPCMFormatFlagIsSignedInteger | kLinearPCMFormatFlagIsBigEndian;
	basic_desc.mChannelsPerFrame = 1;
	basic_desc.mSampleRate = 44100;
	basic_desc.mBitsPerChannel = 16;

	basic_desc.mFramesPerPacket = 1;
	basic_desc.mBytesPerFrame = 2;
	basic_desc.mBytesPerPacket = 2;

	/* And find the default audio output unit */
    desc.componentType = kAudioUnitComponentType;
    desc.componentSubType = kAudioUnitSubType_Output;
    desc.componentManufacturer = kAudioUnitID_DefaultOutput;
    desc.componentFlags = 0;
    desc.componentFlagsMask = 0;
    
    comp = FindNextComponent(NULL, &desc);
    if(comp == NULL) {
        return -1;
    }

	error = OpenAComponent(comp, &outputAU);
	if(error != noErr)	{
		return -1;
	}

	error = AudioUnitInitialize(outputAU);
	if(error != noErr)	{
		return -1;
	}

	error = AudioUnitSetProperty(outputAU, kAudioUnitProperty_StreamFormat, kAudioUnitScope_Input, 0, &basic_desc, sizeof(basic_desc));
	if(error != noErr)	{
		return -1;
	}

	callback.inputProc = sound_callback;
	callback.inputProcRefCon = NULL;

	error = AudioUnitSetProperty(outputAU, kAudioUnitProperty_SetInputCallback, kAudioUnitScope_Input, 0, &callback, sizeof(callback));
	if(error != noErr)	{
		return -1;
	}

	error = AudioOutputUnitStart(outputAU);
	if(error != noErr)	{
		return -1;
	}

	return 0;
}

void sound_shutdown()	{
	OSStatus error = noErr;
	struct AudioUnitInputCallback callback;

	error = AudioOutputUnitStop(outputAU);
	if(error != noErr)
		return;

	callback.inputProc = NULL;
	callback.inputProcRefCon = NULL;

	error = AudioUnitSetProperty(outputAU, kAudioUnitProperty_SetInputCallback, kAudioUnitScope_Input, 0, &callback, sizeof(callback));
	if(error != noErr)
		return;

	error = CloseComponent(outputAU);
	if(error != noErr)
		return;
}
