/*
	This file is part of CrabEmu.

	Copyright (C) 2005 Lawrence Sebald

	CrabEmu is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	CrabEmu is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with CrabEmu; if not, write to the Free Software
	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "sms.h"
#include "smsvdp.h"
#include "smsz80.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* The hardwired palette when used in a TMS9918 mode */
/* These constants are taken from Sean Young's TMS9918A Documentation */
/* This isn't used yet, hence why its not compiled in  */
#if 0
const static sms_vdp_color_t tms9918_pal[16] = {
	/*  A,   R,   G,   B */
	{   0,   0,   0,   0 }, /* Transparent */ 
	{ 255,   0,   0,   0 }, /* Black */
	{ 255,  33, 200,  66 }, /* Medium Green */
	{ 255,  94, 220, 120 }, /* Light Green */
	{ 255,  84,  85, 237 }, /* Dark Blue */
	{ 255, 125, 118, 252 }, /* Light Blue */
	{ 255, 212,  82,  77 }, /* Dark Red */
	{ 255,  66, 235, 245 }, /* Cyan */
	{ 255, 252,  85,  84 }, /* Medium Red */
	{ 255, 255, 121, 120 }, /* Light Red */
	{ 255, 212, 193,  84 }, /* Dark Yellow */
	{ 255, 230, 206, 128 }, /* Light Yellow */
	{ 255,  33, 176,  59 }, /* Dark Green */
	{ 255, 201,  91, 186 }, /* Magenta */
	{ 255, 204, 204, 204 }, /* Gray */
	{ 255, 255, 255, 255 }  /* White */
};
#endif

extern int sms_console;
sms_vdp_t smsvdp;

static void update_local_pal_sms(int num)	{
	/* Calculate the ARGB8888 color from the RGB222 color */
	smsvdp.pal[num].r = (smsvdp.cram[num] & 0x03) << 6;
	smsvdp.pal[num].g = (smsvdp.cram[num] & 0x0C) << 4;
	smsvdp.pal[num].b = (smsvdp.cram[num] & 0x30) << 2;
	smsvdp.pal[num].a = ((num & 0x0F) ? 0xFF : 0x00);
}

static void update_local_pal_gg(int num)	{
	/* Calculate the ARGB8888 color from the RGB444 color */
	smsvdp.pal[num >> 1].r = (smsvdp.cram[num] & 0x0F) << 4;
	smsvdp.pal[num >> 1].g = (smsvdp.cram[num] & 0xF0);
	smsvdp.pal[num >> 1].b = (smsvdp.cram[num + 1] & 0x0F) << 4;
	smsvdp.pal[num >> 1].a = ((num & 0x0F) ? 0xFF : 0x00);
}

/* We will assume NTSC 192-line mode for now */
static inline void update_vcnt(int line)	{
	switch(line)	{
		case 0x00:
			smsvdp.vcnt = 0;
			break;
		case 0xDB:
			smsvdp.vcnt = 0xD5;
			break;
		default:
			++smsvdp.vcnt;
			break;
	}
}

/* ----------------------------------------------- 
   Implementation
   ----------------------------------------------- */

static void sms_vdp_update_cache(int pat)	{
	unsigned char *bitplane;
	int i, j, tmp;
	unsigned char pixel;

	/* Now that we have that, convert rows from bitplane
	   format to paletted using the native palette */
	for(i = 0; i < 8; ++i)	{
		/* Determine where we should start converting
		   Formula: VRAM base + pattern number * 32 + line * 4 */
		bitplane = &smsvdp.vram[(pat << 5) + (i << 2)];

		for(j = 0; j < 8; ++j)	{
			tmp = 7 - j;
			pixel = ((bitplane[0] >> tmp) & 1) | (((bitplane[1] >> tmp) & 1) << 1) |
			        (((bitplane[2] >> tmp) & 1) << 2) | (((bitplane[3] >> tmp) & 1) << 3);
			smsvdp.pattern[pat].texture[(i << 3) + j] = pixel;
		}
	}
}

static void sms_vdp_m4_draw_bg(int line)	{
	int row, col, tile, i, tmp;
	int rendercol, renderrow, xoff, yoff;
	unsigned char *name_table = &smsvdp.vram[(smsvdp.regs[2] & 0x0E) << 10];
	unsigned char *pixels;
	sms_vdp_color_t *pal;

	if(line >= 192)
		/* Don't do anything, we're out of the displayable range */
		return;

	row = line >> 3;

	/* Figure out what column/row to render first */
	if(row < 2 && smsvdp.regs[0] & 0x40)	{
		/* Do not scroll the top two rows, if requested. */
		rendercol = 0;
		xoff = 0;
	}
	else	{
		tmp = (smsvdp.regs[8] & 0xF8) >> 3;
		rendercol = 32 - tmp;

		if(rendercol == 32)
			rendercol = 0;

		/* Our fine scroll value for the X direction */
		xoff = (smsvdp.regs[8] & 0x07);
	}

	tmp = (line + smsvdp.regs[9]) % 224;
	renderrow = (tmp >> 3);

	/* Determine our fine scroll value */
	yoff = (smsvdp.regs[9] & 0x07);

	/* Render each pixel of background data for this line */
	for(col = 0; col < 32; ++col)	{
		/* Get the tile that we're rendering */
		tile = name_table[(renderrow << 6) + (rendercol << 1)] | (name_table[(renderrow << 6) + (rendercol << 1) + 1] << 8);

		/* Check if the pattern is dirtied, if so, clean it up */
		if(smsvdp.pattern[tile & 0x1FF].dirty)
			sms_vdp_update_cache(tile & 0x1FF);

		/* Find the texture */
		switch(tile & 0x400)	{
			case 0x000:
				/* Non-vertically flipped tile */
				pixels = &smsvdp.pattern[tile & 0x1FF].texture[((line + yoff) & 0x07) << 3];
				break;
			case 0x400:
				/* Vertically flipped tile */
				pixels = &smsvdp.pattern[tile & 0x1FF].texture[(7 - ((line + yoff) & 0x07)) << 3];
				break;
			default:
				return;
		}

		/* Which palette are we using? */
		switch(tile & 0x800)	{
			case 0x000:
				/* 1st 16 entries */
				pal = &smsvdp.pal[0];
				break;
			case 0x800:
				/* 2nd 16 entries */
				pal = &smsvdp.pal[16];
				break;
			default:
				return;
		}

		/* Is this a high-priority tile? */
		smsvdp.bg_prio[row][col] = (tile & 0x1000) ? 1 : 0;

		/* Draw the pixels */
		for(i = 0; i < 8; ++i)	{
			if(((col << 3) + i + xoff) > 255)	{
				break;
			}

			switch(tile & 0x200)	{
				case 0x000:
					/* Normal pattern */
					smsvdp.framebuffer[(line << 10) + (((col << 3) + i + xoff) << 2)] = pal[pixels[i]].r;
					smsvdp.framebuffer[(line << 10) + (((col << 3) + i + xoff) << 2) + 1] = pal[pixels[i]].g;
					smsvdp.framebuffer[(line << 10) + (((col << 3) + i + xoff) << 2) + 2] = pal[pixels[i]].b;
					smsvdp.framebuffer[(line << 10) + (((col << 3) + i + xoff) << 2) + 3] = pal[pixels[i]].a;
					break;
				case 0x200:
					/* Horizontally flipped pattern */
					smsvdp.framebuffer[(line << 10) + (((col << 3) + i + xoff) << 2)] = pal[pixels[(7 - i)]].r;
					smsvdp.framebuffer[(line << 10) + (((col << 3) + i + xoff) << 2) + 1] = pal[pixels[(7 - i)]].g;
					smsvdp.framebuffer[(line << 10) + (((col << 3) + i + xoff) << 2) + 2] = pal[pixels[(7 - i)]].b;
					smsvdp.framebuffer[(line << 10) + (((col << 3) + i + xoff) << 2) + 3] = pal[pixels[(7 - i)]].a;
					break;
			}
		}
		rendercol = (rendercol + 1) & 0x1F;
	}

	/* Determine if we should mask column 0, and if so, do it */
	if(smsvdp.regs[0] & 0x20)	{
		for(i = 0; i < 8; ++i)	{
			tmp = (smsvdp.regs[7] & 0x0F) | 0x10;
			smsvdp.framebuffer[(line << 10) + (i << 2)] = smsvdp.pal[tmp].r;
			smsvdp.framebuffer[(line << 10) + (i << 2) + 1] = smsvdp.pal[tmp].g;
			smsvdp.framebuffer[(line << 10) + (i << 2) + 2] = smsvdp.pal[tmp].b;
			smsvdp.framebuffer[(line << 10) + (i << 2) + 3] = smsvdp.pal[tmp].a;
		}
	}
}

static void sms_vdp_m4_draw_spr(int line)	{
	unsigned char *sat;
	unsigned char *pix;
	static unsigned char col_tab[0xFF];
	int height, i, n = 0, j, x, y, num;

	/* We're outside the displayable area, punt */
	if(line >= 192)
		return;

	/* First of all, clear out our colision table */
	memset(col_tab, 0, 0xFF);

	/* Normal sprites are 8 pixels high, so start with that */
	height = 8;

	/* If bit 1 of register 1 is set, then the height is 16 */
	if(smsvdp.regs[1] & 0x02)	{
		height = 16;
	}

	/* Handle double-size sprites here */

	/* Determine the location of the Sprite Attribute Table */
	sat = smsvdp.vram + ((smsvdp.regs[5] & 0x7E) << 7);

	/* Parse the table */
	for(i = 0; i < 64; ++i)	{
		/* Check for the end of list flag */
		if(sat[i] == 0xD0)	{
			break;
		}

		/* Check the position of this sprite */
		if(line > sat[i] && line <= sat[i] + height)	{
			/* If its on this line, make sure that we haven't already rendered 8 sprites */
			if(n < 8)	{
				/* Determine the X position, which is stored in sat[i * 2 + 0x80], and can
				   be changed by bit 3 of register 0, also get the patern number, which
				   is stored in sat[i * 2 + 0x81], and finally get the Y position, which
				   is one more than the value stored in sat[i] */
				x = sat[(i << 1) | 0x80] - (smsvdp.regs[0] & 0x08 ? 8 : 0);
				num = sat[(i << 1) | 0x81];
				y = sat[i] + 1;

				/* Which half of the pattern table are we getting the pattern from? */
				num += (smsvdp.regs[6] & 0x04 ? 256 : 0);

				/* Set our pixel pointer up */
				if(height == 16 && line - y > 7)	{
					if(smsvdp.pattern[num + 1].dirty)
						sms_vdp_update_cache(num + 1);
					pix = &smsvdp.pattern[num + 1].texture[(line - y - 8) << 3];
				}
				else	{
					if(smsvdp.pattern[num].dirty)
						sms_vdp_update_cache(num);
					pix = &smsvdp.pattern[num].texture[(line - y) << 3];
				}

				for(j = 0; j < 8; ++j)	{
					/* Don't draw to a negitive coordinate */
					if(x + j < 0)	{
						continue;
					}
					/* Check for collision, set the flag if there is one */
					if(col_tab[x + j] == 1)	{
						smsvdp.status |= 0x20;
						continue;
					}

					/* Make sure not to tromp on a high-priority background tile */
					if(smsvdp.bg_prio[line >> 3][(x + j) >> 3] == 1 &&
					   smsvdp.framebuffer[(line << 10) + ((x + j) << 2) + 3] == 0xFF)
						continue;

					/* Actually draw the pixel */
					if(pix[j])	{
						smsvdp.framebuffer[(line << 10) + ((x + j) << 2)] = smsvdp.pal[pix[j] | 0x10].r;
						smsvdp.framebuffer[(line << 10) + ((x + j) << 2) + 1] = smsvdp.pal[pix[j] | 0x10].g;
						smsvdp.framebuffer[(line << 10) + ((x + j) << 2) + 2] = smsvdp.pal[pix[j] | 0x10].b;
						smsvdp.framebuffer[(line << 10) + ((x + j) << 2) + 3] = smsvdp.pal[pix[j] | 0x10].a;
						col_tab[x + j] = 1;
					}
				}
				++n;
			}
			else	{
				/* Too many sprites on this line, set the flag, and stop rendering */
				smsvdp.status |= 0x40;
				break;
			}
		}
	}
}

void sms_vdp_execute(int line)	{
	int vdp_mode;

	vdp_mode = ((smsvdp.regs[0x0] & 0x04) << 1) |
		((smsvdp.regs[0x1] & 0x08) >> 1) |
		((smsvdp.regs[0x0] & 0x02)) |
		((smsvdp.regs[0x1] & 0x10) >> 4);

	/* Draw only if the display is enabled */
	if(smsvdp.regs[0x1] & 0x40)	{
		switch(vdp_mode)	{
			case 0x08: /* Mode 4 */
			case 0x0A:
			case 0x0C:
			case 0x0F:
				sms_vdp_m4_draw_bg(line);
				sms_vdp_m4_draw_spr(line);
				break;
			default:
				fprintf(stderr, "sms_vdp_execute: Unsupported mode 0x%02x.\n", vdp_mode);
		}
	}

	update_vcnt(line);

	if(line <= 192)	{
		if(smsvdp.linecnt == 0)	{
			/* The Register 10 line count will underflow, interrupt
			   if enabled */
			smsvdp.linecnt = smsvdp.regs[0xA];
			if(smsvdp.regs[0x0] & 0x10)	{
				sms_z80_irq();
			}
		}
		else	{
			/* Decrement the line counter */
			--smsvdp.linecnt;
		}
	}
	else	{
		/* Reload the line counter */
		smsvdp.linecnt = smsvdp.regs[0xA];
	}

	if(line == 192)	{
		/* Set the vblank flag, and signal an IRQ, if desired */
		smsvdp.status |= 0x80;
		if(smsvdp.regs[0x1] & 0x20)	{
			sms_z80_irq();
		}
	}
}

void sms_vdp_data_write(unsigned char data)	{
	/* Clear the bytes written flag */
	smsvdp.flags &= (~0x01);

	/* First, update the read buffer */
	smsvdp.read_buf = data;

	/* Check the destination of this byte */
	switch(smsvdp.code)	{
		case 0x00:
		case 0x01:
		case 0x02:
			smsvdp.vram[smsvdp.addr] = data;
			smsvdp.pattern[smsvdp.addr >> 5].dirty = 1;
			break;
		case 0x03:
			if(sms_console != CONSOLE_GG)	{
				smsvdp.cram[smsvdp.addr & 0x1F] = data;
				update_local_pal_sms(smsvdp.addr & 0x1F);
			}
			else	{
				if(!(smsvdp.addr & 0x01))	{
					smsvdp.pal_latch = data;
				}
				else	{
					smsvdp.cram[smsvdp.addr & 0x3E] = smsvdp.pal_latch;
					smsvdp.cram[smsvdp.addr & 0x3F] = data;
					update_local_pal_gg(smsvdp.addr & 0x3E);
				}
			}
			break;
		default:
			/* We shouldn't get here */
			fprintf(stderr, "sms_vdp_data_write: Invalid code?!\n");
			return;
	}

	/* Update the address register, and wrap, if needed */
	smsvdp.addr = (smsvdp.addr + 1) & 0x3FFF;
}

void sms_vdp_ctl_write(unsigned char data)	{
	/* First, update the code/address registers */
	if(smsvdp.flags & 0x01)	{
		/* One byte written, write the last one */
		smsvdp.addr = (data & 0x3F) << 8 | (smsvdp.addr_latch & 0xFF);
		smsvdp.code = (data & 0xC0) >> 6;
		smsvdp.flags &= (~0x01);
	}
	else	{
		/* A multiple of 2 bytes written, start over */
		smsvdp.addr = (smsvdp.addr & 0x3F00) | (data);
		smsvdp.addr_latch = data;
		smsvdp.flags |= 0x01;
	}

	/* Now, handle register writes, if needed */
	if(smsvdp.flags == 0 && smsvdp.code == 0x02)	{
		smsvdp.regs[(smsvdp.addr >> 8) & 0x0F] = (smsvdp.addr & 0xFF);
		/* Handle any state changes here */
	}

	/* Code 0 - Read a byte of vram, put it in the buffer, and increment
	   the address (that last part is important!) */
	if(smsvdp.flags == 0 && smsvdp.code == 0x00)	{
		smsvdp.read_buf = smsvdp.vram[smsvdp.addr];
		smsvdp.addr = (smsvdp.addr + 1) & 0x3FFF;
	}
}

unsigned char sms_vdp_hcnt_read(void)	{
	/* This will always return 0, for the time being */
	/* return (unsigned char)(smsvdp.hcnt >> 1);*/
	return (unsigned char) 0;
}

unsigned char sms_vdp_vcnt_read(void)	{
	return smsvdp.vcnt;
}

unsigned char sms_vdp_data_read(void)	{
	unsigned char tmp;

	/* Clear the bytes written flag */
	smsvdp.flags &= (~0x01);

	/* Fetch what is already in the buffer */
	tmp = smsvdp.read_buf;

	/* Fetch a new byte for the buffer */
	smsvdp.read_buf = smsvdp.vram[smsvdp.addr];

	/* And finally, increment the address */
	smsvdp.addr = (smsvdp.addr + 1) & 0x3FFF;

	return tmp;
}

unsigned char sms_vdp_status_read(void)	{
	unsigned char tmp;

	/* Clear the bytes written flag */
	smsvdp.flags &= (~0x01);

	/* Fetch our flags, so that we can clear stale ones */
	tmp = smsvdp.status;
	smsvdp.status &= (~0xE0);

	return tmp;
}

void sms_vdp_dump_vram(const char *fn)	{
	FILE *fp = fopen(fn, "wb");

	if(fp == NULL)
		return;

	fwrite(smsvdp.vram, 0x4000, 1, fp);
	fclose(fp);
}

int sms_vdp_init(void)	{
	int i;

	/* Initialize the VDP emulation to a sane state */
	smsvdp.code = 0;
	smsvdp.addr = 0;
	smsvdp.read_buf = 0;
	smsvdp.status = 0;
	smsvdp.hcnt = 0;
	smsvdp.vcnt = 0;
	smsvdp.flags = 0;
	smsvdp.linecnt = 0xFF;
	smsvdp.pal_latch = 0;

	/* Set some sane register values */
	smsvdp.regs[0x0] = 0x00;
	smsvdp.regs[0x1] = 0x00;
	smsvdp.regs[0x2] = 0x0E;
	smsvdp.regs[0x3] = 0x00;
	smsvdp.regs[0x4] = 0x00;
	smsvdp.regs[0x5] = 0x7F;
	smsvdp.regs[0x6] = 0x00;
	smsvdp.regs[0x7] = 0x00;
	smsvdp.regs[0x8] = 0x00;
	smsvdp.regs[0x9] = 0x00;
	smsvdp.regs[0xA] = 0x00;

	/* Mark all paterns as dirty, and set their offsets */
	for(i = 0; i < 512; ++i)	{
		smsvdp.pattern[i].dirty = 1;
		smsvdp.pattern[i].offset = 32 * i;
		/* There's no use in clearing the textures, since the will be
		   overwritten before they're used, anyway */
	}

	/* Allocate memory */
	smsvdp.cram = (unsigned char *)malloc(64);

	if(smsvdp.cram == NULL)	{
		fprintf(stderr, "sms_vdp_init: Could not allocate CRAM!\n");
		return -1;
	}

	smsvdp.pal = (sms_vdp_color_t *)malloc(sizeof(sms_vdp_color_t) * 32);

	if(smsvdp.pal == NULL)	{
		fprintf(stderr, "sms_vdp_init: Could not allocate local Palette!\n");
		free(smsvdp.cram);
		return -1;
	}

	smsvdp.vram = (unsigned char *)malloc(0x4000);

	if(smsvdp.vram == NULL)	{
		fprintf(stderr, "sms_vdp_init: Could not allocate VRAM!\n");
		free(smsvdp.cram);
		free(smsvdp.pal);
		return -1;
	}

	smsvdp.framebuffer = (unsigned char *)malloc(256 * 256 * 4);

	if(smsvdp.framebuffer == NULL)	{
		fprintf(stderr, "sms_vdp_init: Could not allocate framebuffer!\n");
		free(smsvdp.cram);
		free(smsvdp.pal);
		free(smsvdp.vram);
		return -1;
	}

	return 0;
}

int sms_vdp_shutdown(void)	{
	/* Free all RAM used */
	free(smsvdp.cram);
	free(smsvdp.pal);
	free(smsvdp.vram);
	free(smsvdp.framebuffer);

	return 0;
}
