/*
    This file is part of CrabEmu.

    Copyright (C) 2005, 2006, 2007 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "sms.h"
#include "smsmem.h"
#include "smsvdp.h"
#include "sn76489.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef _arch_dreamcast
#include <zlib/zlib.h>
#include <dc/vmu_pkg.h>
#endif

static unsigned char paging_regs[4];
static unsigned char ram[8 * 1024];

static unsigned char *cart_rom = NULL;
static unsigned long cart_len;
static unsigned char cart_ram[0x8000];
static unsigned char gg_regs[7];
static char *savefn = NULL;
static unsigned char wrote_to_cartram = 0;

static unsigned char *rom_page0;
static unsigned char *rom_page1;
static unsigned char *rom_page2;

static int mapper;

extern unsigned short sms_pad;
extern sn76489_t psg;
extern int sms_console;

static unsigned char *read_map[256];

static void remap_page0()   {
    int i;

    if(paging_regs[1] && cart_len > 0x4000) {
        rom_page0 = &cart_rom[0x4000 * (paging_regs[1] % (cart_len / 0x4000))];
    }
    else    {
        rom_page0 = cart_rom;
    }

    if(mapper != SMS_MAPPER_SEGA)   {
        i = 0x00;
    }
    else    {
        i = 0x04;
        read_map[0] = cart_rom;
        read_map[1] = cart_rom + 0x100;
        read_map[2] = cart_rom + 0x200;
        read_map[3] = cart_rom + 0x300;
    }

    for(; i < 0x40; ++i)    {
        read_map[i] = rom_page0 + (i << 8);
    }
}

static void remap_page1()   {
    int i;

    if(paging_regs[2] && cart_len > 0x4000) {
        rom_page1 = &cart_rom[0x4000 * (paging_regs[2] % (cart_len / 0x4000))];
    }
    else    {
        rom_page1 = cart_rom;
    }

    for(i = 0x40; i < 0x80; ++i)    {
        read_map[i] = rom_page1 + ((i & 0x3F) << 8);
    }
}

static void remap_page2()   {
    int i;

    if(paging_regs[0] & 0x08 && mapper == SMS_MAPPER_SEGA)  {
        /* Cartridge RAM enable */
        if(paging_regs[0] & 0x04)   {
            rom_page2 = &cart_ram[0x4000];
        }
        else    {
            rom_page2 = &cart_ram[0];
        }
    }
    else if(paging_regs[3] && cart_len > 0x4000)    {
        rom_page2 = &cart_rom[0x4000 * (paging_regs[3] % (cart_len / 0x4000))];
    }
    else    {
        rom_page2 = cart_rom;
    }

    for(i = 0x80; i < 0xC0; ++i)    {
        read_map[i] = rom_page2 + ((i & 0x3F) << 8);
    }
}

static void reorganize_pages()  {
    remap_page0();
    remap_page1();
    remap_page2();
}

/* This function is based on code from MEKA, the duty of this function is to
   try to autodetect what mapper type this cartridge should be using. */
static void setup_mapper()  {
    int codemasters, sega, korean;
    unsigned char *ptr;

    if(cart_len < 0x8000)   {
        /* The cartridge contains less than 32KB of data, so there probably
           isn't even a mapper in it, but just to be safe, in case it has
           SRAM for some reason, we'll just assume the Sega Mapper. */
        mapper = SMS_MAPPER_SEGA;
        return;
    }

    sega = 0;
    codemasters = 0;
    korean = 0;
    ptr = cart_rom;

    while(ptr < cart_rom + 0x8000)  {
        if((*ptr++) == 0x32)  {
            unsigned short value = (*ptr) | ((*(ptr + 1)) << 8);

            ptr += 2;

            if(value == 0xFFFF)
                ++sega;
            else if(value == 0x8000)
                ++codemasters;
            else if(value == 0xA000)
                ++korean;
        }
    }

    if(codemasters > sega && codemasters > korean)
        mapper = SMS_MAPPER_CODEMASTERS;
    else if(korean > codemasters && korean > sega)
        mapper = SMS_MAPPER_KOREAN;
    else
        mapper = SMS_MAPPER_SEGA;
}

void sms_mem_write(unsigned short addr, unsigned char data) {
    if(sms_console == CONSOLE_SG1000)   {
        if(addr < 0xC000)   {
            /* Cartridge ROM, write illegal */
#ifdef DEBUG
            fprintf(stdout, "sms_mem_write: Attempt to write to ROM (addr = %04x).\n", addr);
#endif
        }
        else    {
            ram[addr & 0x07FF] = data;
        }
    }
    else if(mapper == SMS_MAPPER_SEGA)  {
        if(addr < 0x8000)   {
            /* Cartridge ROM page 0 and 1, write illegal */
#ifdef DEBUG
            fprintf(stdout, "sms_mem_write: Attempt to write to ROM (addr = %04x).\n", addr);
#endif
        }
        else if(addr < 0xC000)  {
            if(paging_regs[0] & 0x08)   {
                /* Cartridge RAM page 0 or 1 */
                rom_page2[addr & 0x3FFF] = data;
                wrote_to_cartram = 1;
            }
            else    {
                /* Cartridge ROM page 2, write illegal */
#ifdef DEBUG
                fprintf(stdout, "sms_mem_write: Attempt to write to ROM (addr = %04x).\n", addr);
#endif
            }
        }
        else    {
            /* On-board RAM */
            ram[addr & 0x1FFF] = data;

            /* Check if this is bound for a paging register */
            if(addr > 0xFFFB)   {
                if(addr == 0xFFFC)  {
                    paging_regs[0] = data;
                    remap_page2();
                }
                else if(addr == 0xFFFD) {
                    paging_regs[1] = data;
                    remap_page0();
                }
                else if(addr == 0xFFFE) {
                    paging_regs[2] = data;
                    remap_page1();
                }
                else    {
                    paging_regs[3] = data;
                    remap_page2();
                }
            }
        }
    }
    else if(mapper == SMS_MAPPER_CODEMASTERS)   {
        if(addr < 0x4000)   {
            paging_regs[1] = data;
            remap_page0();
        }
        else if(addr < 0x8000)  {
            paging_regs[2] = data;
            remap_page1();
        }
        else if(addr < 0xC000)  {
            paging_regs[3] = data;
            remap_page2();
        }
        else    {
            ram[addr & 0x1FFF] = data;
        }
    }
    else if(mapper == SMS_MAPPER_KOREAN)    {
        if(addr == 0xA000)  {
            paging_regs[3] = data;
            remap_page2();
        }
        else if(addr < 0xC000)   {
            /* Cartridge ROM write illegal */
#ifdef DEBUG
            fprintf(stdout, "sms_mem_write: Attempt to write to ROM (addr = %04x).\n", addr);
#endif
        }
        else    {
            ram[addr & 0x1FFF] = data;
        }
    }
}

void sms_port_write(unsigned short port, unsigned char data)    {
    if(port < 0x07 && sms_console == CONSOLE_GG)    {
        gg_regs[port] = data;

        if(port == 0x06)    {
            psg.output_channels = data;
        }
    }
    else if(port < 0x40)    {
        if(port & 0x01) {
            /* I/O Control register */
        }
        else    {
            /* Memory Control register */
        }
    }
    else if(port < 0x80)    {
        /* SN76489 PSG */
        sn76489_write(&psg, data);
    }
    else if(port < 0xC0)    {
        if(port & 0x01) {
            /* VDP Control port */
            sms_vdp_ctl_write(data);
        }
        else    {
            /* VDP Data port */
            sms_vdp_data_write(data);
        }
    }
    else if(port < 0x100)   {
    }
    else    {
        fprintf(stderr, "sms_port_write: Write to invalid port: %04x\n", port);
    }
}

unsigned char sms_mem_read(unsigned short addr) {
    return read_map[addr >> 8][addr & 0xFF];
}

unsigned char sms_port_read(unsigned short port)    {
    if(port < 0x07 && sms_console == CONSOLE_GG)    {
        return gg_regs[port];
    }
    else if(port < 0x40)    {
        return 0xFF;
    }
    else if(port < 0x80)    {
        if(port & 0x01) {
            return sms_vdp_hcnt_read();
        }
        else    {
            return sms_vdp_vcnt_read();
        }
    }
    else if(port < 0xC0)    {
        if(port & 0x01) {
            return sms_vdp_status_read();
        }
        else    {
            return sms_vdp_data_read();
        }
    }
    else if(port < 0x100)   {
        if(port & 0x01) {
            /* I/O port B/misc register */
            return (sms_pad >> 8) & 0xFF;
        }
        else    {
            /* I/O port A/B register */
            return (sms_pad & 0xFF);
        }
    }
    else    {
        fprintf(stderr, "sms_read_port: Attempt to read invalid port: %04x\n", port);
        return 0;
    }
}

#ifndef _arch_dreamcast

int sms_write_cartram_to_file(void) {
    FILE *fp;

    if(savefn == NULL)
        return -1;

    if(!wrote_to_cartram)
        return 0;

    fp = fopen(savefn, "wb");
    if(!fp)
        return -1;

    fwrite(cart_ram, 0x8000, 1, fp);
    fclose(fp);

    wrote_to_cartram = 0;

    return 0;
}

int sms_read_cartram_from_file(void)    {
    FILE * fp;

    fp = fopen(savefn, "rb");
    
    if(fp != NULL)  {
        fread(cart_ram, 0x8000, 1, fp);
        fclose(fp);

        return 0;
    }

    return -1;
}

#else

int sms_write_cartram_to_file(void) {
    vmu_pkg_t pkg;
    uint8 *pkg_out;
    uint8 *comp;
    int pkg_size;
    uint32 len;
    FILE *fp;
    char prodcode[7];
    char savename[15];

    if(savefn == NULL)
        return -1;

    if(!wrote_to_cartram)
        return 0;

    sprintf(prodcode, "%d%d%d%d%d", (cart_rom[0x7FFE] & 0xF0) >> 4,
            (cart_rom[0x7FFD] & 0xF0) >> 4, (cart_rom[0x7FFD] & 0x0F),
            (cart_rom[0x7FFC] & 0xF0) >> 4, (cart_rom[0x7FFC] & 0x0F));

    sprintf(savename, "/vmu/a1/%s", prodcode);

    comp = (uint8 *)malloc(0x8000);
    compress2(comp, &len, cart_ram, 0x8000, 9);

    strcpy(pkg.desc_short, "CrabEmu Save");
    strcpy(pkg.desc_long, prodcode);
    strcpy(pkg.app_id, "CrabEmu");
    pkg.icon_cnt = 0;
    pkg.icon_anim_speed = 0;
    memset(pkg.icon_pal, 0, 32);
    pkg.icon_data = NULL;
    pkg.eyecatch_type = VMUPKG_EC_NONE;
    pkg.data_len = len;
    pkg.data = comp;

    vmu_pkg_build(&pkg, &pkg_out, &pkg_size);

    fp = fopen(savename, "w");

    if(!fp)
        return -1;

    fwrite(pkg_out, pkg_size, 1, fp);
    fclose(fp);

    free(pkg_out);
    free(comp);

    wrote_to_cartram = 0;

    return 0;
}

int sms_read_cartram_from_file(void)    {
    vmu_pkg_t pkg;
    uint8 *pkg_out;
    int pkg_size;
    char prodcode[7];
    FILE *fp;
    char savename[15];
    uint32 real_size = 0x8000;

    if(savefn == NULL)
        return -1;

    if(wrote_to_cartram)
        return -1;

    sprintf(prodcode, "%d%d%d%d%d", (cart_rom[0x7FFE] & 0xF0) >> 4,
            (cart_rom[0x7FFD] & 0xF0) >> 4, (cart_rom[0x7FFD] & 0x0F),
            (cart_rom[0x7FFC] & 0xF0) >> 4, (cart_rom[0x7FFC] & 0x0F));
    
    sprintf(savename, "/vmu/a1/%s", prodcode);

    fp = fopen(savename, "r");

    if(!fp)
        return -1;

    fseek(fp, 0, SEEK_SET);
    fseek(fp, 0, SEEK_END);
    pkg_size = ftell(fp);
    fseek(fp, 0, SEEK_SET);

    pkg_out = (uint8 *)malloc(pkg_size);
    fread(pkg_out, pkg_size, 1, fp);
    fclose(fp);

    vmu_pkg_parse(pkg_out, &pkg);

    uncompress(cart_ram, &real_size, pkg.data, pkg.data_len);

    free(pkg_out);

    return 0;
}

#endif /* _arch_dreamcast */

int sms_mem_load_rom(const char *fn)    {
    FILE *fp;
    int size;
    char *ext;

    fp = fopen(fn, "rb");
    if(!fp) {
        fprintf(stderr, "sms_mem_load_rom: Could not open ROM: %s!\n", fn);
        return -1;
    }

    /* First, determine which type of ROM it is */
    ext = strrchr(fn, '.');
    if(!strcmp(ext, ".gg")) {
        /* GameGear rom */
        sms_console = CONSOLE_GG;
        gui_set_viewport(160, 144);
    }
    else if(!strcmp(ext, ".sc"))    {
        /* SC-3000 rom */
        sms_console = CONSOLE_SG1000;
        gui_set_viewport(256, 192);
    }
    else if(!strcmp(ext, ".sg"))    {
        /* SG-1000 rom */
        sms_console = CONSOLE_SG1000;
        gui_set_viewport(256, 192);
    }
    else    {
        /* Master System rom, hopefully */
        sms_console = CONSOLE_SMS;
        gui_set_viewport(256, 192);
    }

    /* Determine the size of the file */
    fseek(fp, 0, SEEK_END);
    size = ftell(fp);
    fseek(fp, 0, SEEK_SET);

    /* Determine if header/footer info is present */
    if((size % 0x4000) == 512)  {
        /* SMD header present, skip it */
        fseek(fp, 512, SEEK_SET);
        size -= 512;
    }
    else if((size % 0x4000) == 64)  {
        /* Footer present, truncate it */
        size -= 64;
    }

    cart_rom = (unsigned char *)malloc(size);
    fread(cart_rom, size, 1, fp);
    fclose(fp);

    if(savefn)  {
        sms_write_cartram_to_file();
        free(savefn);
        memset(cart_ram, 0, 0x8000);
    }

    savefn = (char *) malloc(strlen(fn) + 5);
    wrote_to_cartram = 0;

    sprintf(savefn, "%s.ces", fn);

    sms_read_cartram_from_file();

    cart_len = size;

    /* Determine which mapper should be used, and set up the paging
       accordingly. */
    mapper = SMS_MAPPER_SEGA;

    if(sms_console != CONSOLE_SG1000)   {
        setup_mapper();
    }

    reorganize_pages();

    return 0;
}

void sms_mem_write_context(FILE *fp)    {
    fwrite(paging_regs, 4, 1, fp);
    fwrite(gg_regs, 7, 1, fp);
    fwrite(ram, 0x2000, 1, fp);
    fwrite(cart_ram, 0x8000, 1, fp);
}

void sms_mem_read_context(FILE *fp) {
    fread(paging_regs, 4, 1, fp);
    fread(gg_regs, 7, 1, fp);
    fread(ram, 0x2000, 1, fp);
    fread(cart_ram, 0x8000, 1, fp);
    reorganize_pages();
}

int sms_mem_init(void)  {
    int i;

    cart_rom = NULL;
    rom_page0 = NULL;
    rom_page1 = NULL;
    rom_page2 = NULL;

    paging_regs[0] = 0;
    paging_regs[1] = 0;
    paging_regs[2] = 1;
    paging_regs[3] = 0;

    gg_regs[0] = 0xC0;
    gg_regs[1] = 0x7F;
    gg_regs[2] = 0xFF;
    gg_regs[3] = 0x00;
    gg_regs[4] = 0xFF;
    gg_regs[5] = 0x00;
    gg_regs[6] = 0xFF;

    for(i = 0xC0; i < 0xE0; ++i)    {
        read_map[i] = ram + ((i - 0xC0) << 8);
    }

    for(i = 0xE0; i < 0x100; ++i)   {
        read_map[i] = ram + ((i - 0xE0) << 8);
    }

    return 0;
}

int sms_mem_shutdown(void)  {
    if(cart_rom != NULL)
        free(cart_rom);

    return 0;
}

void sms_mem_reset(void)    {
    paging_regs[0] = 0;
    paging_regs[1] = 0;
    paging_regs[2] = 1;
    paging_regs[3] = 2;

    gg_regs[0] = 0xC0;
    gg_regs[1] = 0x7F;
    gg_regs[2] = 0xFF;
    gg_regs[3] = 0x00;
    gg_regs[4] = 0xFF;
    gg_regs[5] = 0x00;
    gg_regs[6] = 0xFF;

    reorganize_pages();
}
