/*
    This file is part of CrabEmu.

    Copyright (C) 2007 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdlib.h>
#include <stdio.h>

#include <dc/pvr.h>
#include <dc/maple.h>
#include <dc/maple/controller.h>
#include <arch/arch.h>

#include "sms.h"
#include "smsvdp.h"
#include "smsz80.h"
#include "sn76489.h"
#include "smsmem.h"
#include "sound.h"

pvr_ptr_t tex[2];
pvr_ptr_t txr;
int active_tex;

extern unsigned short sms_pad;
extern const char *menu(const char *dir);

#define UNUSED __attribute__((unused))

/* This function overrides the pvr_txr_load function in KOS to force the
   texture load to take place via the 64bit interface to VRAM. */
#define PVR_LMMODE0 ((int *)0xA05F6884)
void pvr_txr_load_ta(void *src, pvr_ptr_t dst, uint32 count)    {
    int old_mode = *PVR_LMMODE0;
    pvr_ptr_t real_dst = (pvr_ptr_t)(0x11000000 + (((uint32) dst) & 0x7FFFFF));

    *PVR_LMMODE0 = 0;
    pvr_txr_load(src, real_dst, count);
    *PVR_LMMODE0 = old_mode;
}

void render_scene(void) {
    pvr_vertex_t vert;
    pvr_poly_cxt_t cxt;
    pvr_poly_hdr_t hdr;
    float y = (480.0f - (smsvdp.lines << 1)) / 2.0f;

    pvr_txr_load_ta(smsvdp.framebuffer, txr, 256 * 256 * 2);
    active_tex ^= 1;
    txr = tex[active_tex];

    pvr_poly_cxt_txr(&cxt, PVR_LIST_OP_POLY,
                     PVR_TXRFMT_RGB565 | PVR_TXRFMT_NONTWIDDLED, 256, 256,
                     txr, PVR_FILTER_NONE);
    pvr_poly_compile(&hdr, &cxt);

    pvr_wait_ready();
    pvr_scene_begin();

    pvr_list_begin(PVR_LIST_OP_POLY);

    pvr_prim(&hdr, sizeof(pvr_poly_hdr_t));

    vert.flags = PVR_CMD_VERTEX;
    vert.argb = 0xFFFFFFFF;
    vert.oargb = 0;

    vert.x = 64.0f;
    vert.y = y + (smsvdp.lines << 1);
    vert.z = 1.0f;
    vert.u = 0.0f;
    vert.v = smsvdp.lines / 256.0f;
    pvr_prim(&vert, sizeof(pvr_vertex_t));

    vert.y = y;
    vert.v = 0.0f;
    pvr_prim(&vert, sizeof(pvr_vertex_t));

    vert.x = 576.0f;
    vert.y = y + (smsvdp.lines << 1);
    vert.u = 1.0f;
    vert.v = smsvdp.lines / 256.0f;
    pvr_prim(&vert, sizeof(pvr_vertex_t));

    vert.flags = PVR_CMD_VERTEX_EOL;
    vert.y = y;
    vert.v = 0.0f;
    pvr_prim(&vert, sizeof(pvr_vertex_t));

    pvr_list_finish();
    pvr_scene_finish();
}

void emu_loop(void) {
    maple_device_t *cont1, *cont2;
    cont_state_t *state;
    uint16 mask;

#ifdef TIMING_TEST
    int frames = 0;
    time_t thetime = time(NULL);
#endif

    cont1 = maple_enum_type(0, MAPLE_FUNC_CONTROLLER);
    cont2 = maple_enum_type(1, MAPLE_FUNC_CONTROLLER);

    for(;;) {
        sms_frame();
        render_scene();

        mask = 0x0000;

        if(cont1)   {
            state = (cont_state_t *)maple_dev_status(cont1);
            if(state)   {
                if(state->buttons & CONT_A)
                    mask |= SMS_PAD1_A;
                if(state->buttons & CONT_B)
                    mask |= SMS_PAD1_B;
                if(state->buttons & CONT_DPAD_UP)
                    mask |= SMS_PAD1_UP;
                if(state->buttons & CONT_DPAD_DOWN)
                    mask |= SMS_PAD1_DOWN;
                if(state->buttons & CONT_DPAD_LEFT)
                    mask |= SMS_PAD1_LEFT;
                if(state->buttons & CONT_DPAD_RIGHT)
                    mask |= SMS_PAD1_RIGHT;

                if(state->buttons & CONT_START)
                    sms_button_pressed(GG_START);
                else
                    sms_button_released(GG_START);

                if(state->rtrig > 20)
                    sms_write_cartram_to_file();

                if(state->ltrig > 20)   {
                    sound_reset_buffer();
                    sms_reset();
                    sms_mem_load_rom(menu("/cd/crabemu/roms"));
                }
            }
        }

        if(cont2)   {
            state = (cont_state_t *)maple_dev_status(cont2);
            if(state)   {
                if(state->buttons & CONT_A)
                    mask |= SMS_PAD2_A;
                if(state->buttons & CONT_B)
                    mask |= SMS_PAD2_B;
                if(state->buttons & CONT_DPAD_UP)
                    mask |= SMS_PAD2_UP;
                if(state->buttons & CONT_DPAD_DOWN)
                    mask |= SMS_PAD2_DOWN;
                if(state->buttons & CONT_DPAD_LEFT)
                    mask |= SMS_PAD2_LEFT;
                if(state->buttons & CONT_DPAD_RIGHT)
                    mask |= SMS_PAD2_RIGHT;
            }
        }

        sms_pad = mask ^ 0xFFFF;

#ifdef TIMING_TEST
        ++frames;

        if(thetime + 5 <= time(NULL))   {
            printf("%d frames in %d seconds: %f FPS\n", frames, time(NULL) - thetime,
                   (float) frames / (time(NULL) - thetime));
            frames = 0;
            thetime = time(NULL);
        }
#endif
    }
}

void gui_set_viewport(int w UNUSED, int h UNUSED)   {
    /* XXXX */
}

void gui_resized()  {
    /* XXXX */
}

void cont_callback(uint8 addr UNUSED, uint32 buttons UNUSED)    {
    exit(0);
}

int main(int argc UNUSED, char *argv[] UNUSED)  {
    printf("...\n");
    pvr_init_defaults();

    cont_btn_callback(0, CONT_A | CONT_B | CONT_X | CONT_Y, cont_callback);

    tex[0] = pvr_mem_malloc(256 * 256 * 2);
    tex[1] = pvr_mem_malloc(256 * 256 * 2);

    active_tex = 0;
    txr = tex[0];

    sms_init(SMS_VIDEO_NTSC);
    sms_mem_load_rom(menu("/cd/crabemu/roms"));

    emu_loop();
    return 0;
}
