/*
    This file is part of CrabEmu.

    Copyright (C) 2007, 2008 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/* The sound code found herein is roughly based upon code from Genesis Plus/DC,
   which was ported to the Dreamcast by BlackAura (with subsequent modifications
   by Quzar). The sound stuff in GPDC though is pretty much the same as the
   snd_stream stuff in KOS, so I guess Dan Potter should get some credit for
   this code, eh? */

#include <string.h>
#include <dc/spu.h>
#include <dc/g2bus.h>
#include <dc/flashrom.h>
#include <dc/sound/sound.h>

#include "sound.h"
#include <arm/aica_cmd_iface.h>

static const int NTSC_SAMPLES_PER_FRAME = 735;
static const int FRAMES_TO_BUFFER = 12;
static const int SAMPLE_RATE = 44100;

static int initted = 0;
static uint32 ptr[2];
static int buf_pos;

/* Pretty much the same as dcsnd_stop() in GPDC */
static void sound_stop(void)    {
    AICA_CMDSTR_CHANNEL(tmp, cmd, chan);

    if(!initted)
        return;

    /* Reset the left and right channels */
    cmd->cmd = AICA_CMD_CHAN;
    cmd->timestamp = 0;
    cmd->size = AICA_CMDSTR_CHANNEL_SIZE;
    cmd->cmd_id = 0;
    chan->cmd = AICA_CH_CMD_STOP;
    snd_sh4_to_aica(tmp, cmd->size);

#if CRABEMU_DC_ENABLE_STEREO
    cmd->cmd_id = 1;
    snd_sh4_to_aica(tmp, cmd->size);
#endif

    /* Reset the sound buffers */
    spu_memset(ptr[0], 0, NTSC_SAMPLES_PER_FRAME * FRAMES_TO_BUFFER * 2);
    spu_memset(ptr[1], 0, NTSC_SAMPLES_PER_FRAME * FRAMES_TO_BUFFER * 2);
}

/* This function shamelessly mostly copied from GPDC's dcsnd_play() function.
   Note though, the special handling for a system set to mono mode is gone.
   No Master System games can do stereo sound, and very few Game Gear games
   really seem to take advantage of it (probably because the system itself
   only had one speaker). */
static void sound_start(void)   {
    AICA_CMDSTR_CHANNEL(tmp, cmd, chan);

	if(!initted)
		return;

	/* Make sure these are sync'd (and/or delayed) */
	snd_sh4_to_aica_stop();

	/* Set up channel 0 */
	cmd->cmd = AICA_CMD_CHAN;
	cmd->timestamp = 0;
	cmd->size = AICA_CMDSTR_CHANNEL_SIZE;
	cmd->cmd_id = 0;
	chan->cmd = AICA_CH_CMD_START | AICA_CH_START_DELAY;
	chan->base = ptr[0];
	chan->type = AICA_SM_16BIT;
	chan->length = NTSC_SAMPLES_PER_FRAME * FRAMES_TO_BUFFER;
	chan->loop = 1;
	chan->loopstart = 0;
	chan->loopend = NTSC_SAMPLES_PER_FRAME * FRAMES_TO_BUFFER;
	chan->freq = SAMPLE_RATE;
	chan->vol = 255;
	chan->pan = 128;
	snd_sh4_to_aica(tmp, cmd->size);

#ifdef CRABEMU_DC_ENABLE_STEREO
	/* Set up channel 1 */
	cmd->cmd_id = 1;
	chan->base = ptr[1];
	chan->pan = 255;
	snd_sh4_to_aica(tmp, cmd->size);

	/* Send synchronised start command */
	cmd->cmd_id = (1 << 0) | (1 << 1);
#else
    cmd->cmd_id = 1;
#endif
	chan->cmd = AICA_CH_CMD_START | AICA_CH_START_SYNC;
	snd_sh4_to_aica(tmp, cmd->size);

	/* Set up streaming state */
	buf_pos = 8 * NTSC_SAMPLES_PER_FRAME * 2;

	/* Start both sound buffers playing */
	snd_sh4_to_aica_start();
}

void sound_update_buffer_noint(signed short *bufl, signed short *bufr,
                               int length)  {
    spu_memload(ptr[0] + buf_pos, bufl, length);
#ifdef CRABEMU_DC_ENABLE_STEREO
    spu_memload(ptr[1] + buf_pos, bufr, length);
#endif
    buf_pos += length;

    if(buf_pos == NTSC_SAMPLES_PER_FRAME * FRAMES_TO_BUFFER * 2)
        buf_pos = 0;
}

int sound_init(void)    {
    if(initted)
        return 0;

    snd_init();

    /* Allocate the SPU RAM buffers and clear them out */
    ptr[0] = snd_mem_malloc(NTSC_SAMPLES_PER_FRAME * FRAMES_TO_BUFFER * 2);
    ptr[1] = snd_mem_malloc(NTSC_SAMPLES_PER_FRAME * FRAMES_TO_BUFFER * 2);

    spu_memset(ptr[0], 0x00, NTSC_SAMPLES_PER_FRAME * FRAMES_TO_BUFFER * 2);
    spu_memset(ptr[1], 0x00, NTSC_SAMPLES_PER_FRAME * FRAMES_TO_BUFFER * 2);

    initted = 1;

    sound_start();
    
    return 0;
}

void sound_shutdown(void)   {
    sound_stop();

    snd_mem_free(ptr[1]);
    snd_mem_free(ptr[0]);

    initted = 0;
}

void sound_reset_buffer(void)   {
    sound_stop();
    spu_memset(ptr[0], 0x00, NTSC_SAMPLES_PER_FRAME * FRAMES_TO_BUFFER * 2);
    spu_memset(ptr[1], 0x00, NTSC_SAMPLES_PER_FRAME * FRAMES_TO_BUFFER * 2);
    sound_start();
}
