/*
    This file is part of CrabEmu.

    Copyright (C) 2005, 2006, 2007, 2008 Lawrence Sebald

    CrabEmu is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2 
    as published by the Free Software Foundation.

    CrabEmu is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with CrabEmu; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "sms.h"
#include "smsmem.h"
#include "smsvdp.h"
#include "sn76489.h"
#include "93c46.h"
#include "smsz80.h"
#include "mappers.h"
#include "terebi.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef _arch_dreamcast
#include <zlib/zlib.h>
#include <dc/vmu_pkg.h>
#include "icon.h"
#endif

static unsigned char paging_regs[4];
static unsigned char ram[8 * 1024];

static unsigned char *cart_rom = NULL;
static unsigned long cart_len;
static unsigned char cart_ram[0x8000];
static unsigned char gg_regs[7];
static char *savefn = NULL;
static int cartram_enabled = 0;

static unsigned char *rom_page0;
static unsigned char *rom_page1;
static unsigned char *rom_page2;

static int mapper;

extern unsigned short sms_pad;
extern sn76489_t psg;
extern int sms_console;
extern eeprom93c46_t e93c46;

unsigned char *read_map[256];
unsigned char *write_map[256];
static unsigned char dummy_area[256];

typedef void (*remap_page_func)();
static remap_page_func remap_page[4];

static void remap_page0()   {
    int i;

    if(paging_regs[1] && cart_len > 0x4000) {
        rom_page0 = &cart_rom[0x4000 * (paging_regs[1] % (cart_len / 0x4000))];
    }
    else    {
        rom_page0 = cart_rom;
    }

    if(mapper != SMS_MAPPER_SEGA)   {
        i = 0x00;
    }
    else    {
        i = 0x04;
        read_map[0] = cart_rom;
        read_map[1] = cart_rom + 0x100;
        read_map[2] = cart_rom + 0x200;
        read_map[3] = cart_rom + 0x300;
        write_map[0] = dummy_area;
        write_map[1] = dummy_area;
        write_map[2] = dummy_area;
        write_map[3] = dummy_area;
    }

    for(; i < 0x40; ++i)    {
        read_map[i] = rom_page0 + (i << 8);
        write_map[i] = dummy_area;
    }

    sms_z80_set_readmap(read_map);
}

static void remap_page1()   {
    int i;

    if(paging_regs[2] && cart_len > 0x4000) {
        rom_page1 = &cart_rom[0x4000 * (paging_regs[2] % (cart_len / 0x4000))];
    }
    else    {
        rom_page1 = cart_rom;
    }

    for(i = 0x40; i < 0x80; ++i)    {
        read_map[i] = rom_page1 + ((i & 0x3F) << 8);
        write_map[i] = dummy_area;
    }

    sms_z80_set_readmap(read_map);
}

static void remap_page1_codemasters()   {
    int i;

    if((paging_regs[2] & 0x7F) && cart_len > 0x4000) {
        rom_page1 = &cart_rom[0x4000 *
            ((paging_regs[2] & 0x7F) % (cart_len / 0x4000))];
    }
    else    {
        rom_page1 = cart_rom;
    }

    for(i = 0x40; i < 0x80; ++i)    {
        read_map[i] = rom_page1 + ((i & 0x3F) << 8);
        write_map[i] = dummy_area;
    }
    sms_z80_set_readmap(read_map);
}

static void remap_page2()   {
    int i, isram = 0;

    if(mapper == SMS_MAPPER_CASTLE) {
        rom_page2 = cart_ram;

        for(i = 0x80; i < 0xA0; ++i)    {
            write_map[i] = rom_page2 + ((i & 0x3F) << 8);
            read_map[i] = rom_page2 + ((i & 0x3F) << 8);
        }

        for(; i < 0xC0; ++i)    {
            read_map[i] = dummy_area;
        }

        sms_z80_set_readmap(read_map);
        return;
    }
    else if(paging_regs[0] & 0x08 && mapper == SMS_MAPPER_SEGA)  {
        /* Cartridge RAM enable */
        if(paging_regs[0] & 0x04)   {
            rom_page2 = &cart_ram[0x4000];
        }
        else    {
            rom_page2 = &cart_ram[0];
        }

        isram = 1;
        cartram_enabled = 1;
    }
    else if(paging_regs[3] && cart_len > 0x4000)    {
        rom_page2 = &cart_rom[0x4000 * (paging_regs[3] % (cart_len / 0x4000))];
    }
    else    {
        rom_page2 = cart_rom;
    }

    if(isram)   {
        for(i = 0x80; i < 0xC0; ++i)    {
            read_map[i] = rom_page2 + ((i & 0x3F) << 8);
            write_map[i] = rom_page2 + ((i & 0x3F) << 8);
        }
    }
    else    {
        for(i = 0x80; i < 0xC0; ++i)    {
            read_map[i] = rom_page2 + ((i & 0x3F) << 8);
            write_map[i] = dummy_area;
        }
    }

    sms_z80_set_readmap(read_map);
}

static void remap_page2_codemasters()   {
    int i;

    if(paging_regs[3] && cart_len > 0x4000) {
        rom_page2 = &cart_rom[0x4000 * (paging_regs[3] % (cart_len / 0x4000))];
    }
    else    {
        rom_page2 = cart_rom;
    }

    if(paging_regs[2] & 0x80)   {
        for(i = 0x80; i < 0xA0; ++i)    {
            read_map[i] = rom_page2 + ((i & 0x3F) << 8);
            write_map[i] = dummy_area;
        }

        for(; i < 0xC0; ++i)    {
            read_map[i] = cart_ram + ((i & 0x1F) << 8);
            write_map[i] = cart_ram + ((i & 0x1F) << 8);
        }
    }
    else    {
        for(i = 0x80; i < 0xC0; ++i)    {
            read_map[i] = rom_page2 + ((i & 0x3F) << 8);
            write_map[i] = dummy_area;
        }
    }

    sms_z80_set_readmap(read_map);
}

static void reorganize_pages()  {
    remap_page0();
    remap_page1();
    remap_page2();
}

/* This function is based on code from MEKA, the duty of this function is to
   try to autodetect what mapper type this cartridge should be using. */
static void setup_mapper()  {
    int codemasters, sega, korean;
    unsigned char *ptr;

    if(cart_len < 0x8000)   {
        /* The cartridge contains less than 32KB of data, so there probably
           isn't even a mapper in it, but just to be safe, in case it has
           SRAM for some reason, we'll just assume the Sega Mapper. */
        mapper = SMS_MAPPER_SEGA;
        return;
    }

    sega = 0;
    codemasters = 0;
    korean = 0;
    ptr = cart_rom;

    while(ptr < cart_rom + 0x8000)  {
        if((*ptr++) == 0x32)  {
            unsigned short value = (*ptr) | ((*(ptr + 1)) << 8);

            ptr += 2;

            if(value == 0xFFFF)
                ++sega;
            else if(value == 0x8000 || value == 0x4000)
                ++codemasters;
            else if(value == 0xA000)
                ++korean;
        }
    }

    if(codemasters > sega && codemasters > korean)
        mapper = SMS_MAPPER_CODEMASTERS;
    else if(korean > codemasters && korean > sega)
        mapper = SMS_MAPPER_KOREAN;
    else
        mapper = SMS_MAPPER_SEGA;
}

static unsigned char normal_mread(unsigned short addr)  {
    return read_map[addr >> 8][addr & 0xFF];
}

static unsigned char eeprom_mread(unsigned short addr)  {
    if(e93c46.enabled && (addr >> 12) == 8) {
        if(addr == 0x8000)  {
            return eeprom93c46_read();
        }
        else if(addr >= 0x8008 && addr < 0x8088)    {
            int a = (addr - 0x8008) >> 1;

            if(addr & 0x01) {
                return e93c46.data[a] >> 8;
            }
            else    {
                return e93c46.data[a] & 0xFF;
            }
        }
    }

    return read_map[addr >> 8][addr & 0xFF];
}

static void sega_mwrite(unsigned short addr, unsigned char data)    {
    write_map[addr >> 8][addr & 0xFF] = data;

    /* Check if this is bound for a paging register */
    if(addr > 0xFFFB)   {
        if(paging_regs[addr - 0xFFFC] != data)  {
            paging_regs[addr - 0xFFFC] = data;
            remap_page[addr - 0xFFFC]();
        }
    }
}

static void sg1000_mwrite(unsigned short addr, unsigned char data)  {
    write_map[addr >> 8][addr & 0xFF] = data;
}

static void codemasters_mwrite(unsigned short addr, unsigned char data) {
    write_map[addr >> 8][addr & 0xFF] = data;

    if(addr == 0x0000 && paging_regs[1] != data)    {
        paging_regs[1] = data;
        remap_page0();
    }
    else if(addr == 0x4000 && paging_regs[2] != data)   {
        paging_regs[2] = data;
        remap_page1_codemasters();
        remap_page2_codemasters();
    }
    else if(addr == 0x8000 && paging_regs[3] != data)   {
        paging_regs[3] = data;
        remap_page2_codemasters();
    }
}

static void korean_mwrite(unsigned short addr, unsigned char data)  {
    write_map[addr >> 8][addr & 0xFF] = data;

    if(addr == 0xA000 && paging_regs[3] != data)    {
        paging_regs[3] = data;
        remap_page2();
    }
}

static void eeprom_mwrite(unsigned short addr, unsigned char data)  {
    write_map[addr >> 8][addr & 0xFF] = data;

    if(addr == 0x8000 && e93c46.enabled)    {
        eeprom93c46_write(data);
    }
    else if(addr >= 0x8008 && addr < 0x8088 && e93c46.enabled)  {
        int a = (addr - 0x8008) >> 1;

        if(addr & 0x01) {
            e93c46.data[a] &= 0x00FF;
            e93c46.data[a] |= data << 8;
        }
        else    {
            e93c46.data[a] &= 0xFF00;
            e93c46.data[a] |= (data);
        }
    }
    else if(addr == 0xFFFC) {
        eeprom93c46_ctl_write(data);
    }
    else if(addr > 0xFFFC)  {
        /* This is bound for a paging register */
        if(paging_regs[addr - 0xFFFC] != data)  {
            paging_regs[addr - 0xFFFC] = data;
            remap_page[addr - 0xFFFC]();
        }
    }
}

static unsigned short normal_mread16(unsigned short addr)   {
    int top = addr >> 8, bot = addr & 0xFF;
    unsigned short data = read_map[top][bot++];

    if(bot <= 0xFF)
        data |= (read_map[top][bot] << 8);
    else
        data |= (read_map[(unsigned char)(top + 1)][0] << 8);

    return data;
}

static unsigned short eeprom_mread16(unsigned short addr)   {
    int top = addr >> 8, bot = addr & 0xFF;
    unsigned short data;

    if(e93c46.enabled && (addr >> 12) == 8) {
        if(addr == 0x8000)  {
            return eeprom93c46_read() | (read_map[0x80][0x01] << 8);
        }
        else if(addr == 0x7FFF) {
            return (eeprom93c46_read() << 8) | read_map[0x7F][0xFF];
        }
        else if(addr >= 0x8008 && addr < 0x8088)    {
            int a = (addr - 0x8008) >> 1;
            
            if(addr & 0x01) {                
                if(a < 0x7F)    {
                    return (e93c46.data[a] << 8) | (e93c46.data[a + 1] & 0xFF);
                }
                else    {
                    return (e93c46.data[a] << 8) | read_map[0x80][0x88];
                }
            }
            else    {
                return e93c46.data[a];
            }
        }
    }

    data = read_map[top][bot++];

    if(bot <= 0xFF)
        return data | (read_map[top][bot] << 8);
    else
        return data | (read_map[(unsigned char)(top + 1)][0] << 8);
}

static void sega_mwrite16(unsigned short addr, unsigned short data) {
    int top = addr >> 8, bot = addr & 0xFF;

    write_map[top][bot++] = (unsigned char)data;

    if(bot <= 0xFF)
        write_map[top][bot] = (unsigned char)(data >> 8);
    else
        write_map[(unsigned char)(top + 1)][0] = (unsigned char)(data >> 8);

    /* Check if this is bound for a paging register */
    if(addr > 0xFFFB)   {
        paging_regs[addr - 0xFFFC] = (unsigned char)data;
        remap_page[addr - 0xFFFC]();
    }

    if(addr + 1 > 0xFFFB)   {
        paging_regs[addr - 0xFFFC + 1] = (unsigned char)(data >> 8);
        remap_page[addr - 0xFFFC + 1]();
    }
}

static void sg1000_mwrite16(unsigned short addr, unsigned short data)   {
    int top = addr >> 8, bot = addr & 0xFF;

    write_map[top][bot++] = (unsigned char)data;

    if(bot <= 0xFF)
        write_map[top][bot] = (unsigned char)(data >> 8);
    else
        write_map[(unsigned char)(top + 1)][0] = (unsigned char)(data >> 8);
}

static void codemasters_mwrite16(unsigned short addr, unsigned short data)  {
    int top = addr >> 8, bot = addr & 0xFF;

    write_map[top][bot++] = (unsigned char)data;

    if(bot <= 0xFF)
        write_map[top][bot] = (unsigned char)(data >> 8);
    else
        write_map[(unsigned char)(top + 1)][0] = (unsigned char)(data >> 8);

    if(addr == 0x0000)  {
        paging_regs[1] = (unsigned char)data;
        remap_page0();
    }
    else if(addr == 0xFFFF) {
        paging_regs[1] = data >> 8;
        remap_page0();
    }
    else if(addr == 0x4000) {
        paging_regs[2] = (unsigned char)data;
        remap_page1_codemasters();
        remap_page2_codemasters();
    }
    else if(addr == 0x3FFF) {
        paging_regs[2] = data >> 8;
        remap_page1_codemasters();
        remap_page2_codemasters();
    }
    else if(addr == 0x8000) {
        paging_regs[3] = (unsigned char)data;
        remap_page2_codemasters();
    }
    else if(addr == 0x7FFF) {
        paging_regs[3] = data >> 8;
        remap_page2_codemasters();
    }
}

static void korean_mwrite16(unsigned short addr, unsigned short data)   {
    int top = addr >> 8, bot = addr & 0xFF;

    write_map[top][bot++] = (unsigned char)data;

    if(bot <= 0xFF)
        write_map[top][bot] = (unsigned char)(data >> 8);
    else
        write_map[(unsigned char)(top + 1)][0] = (unsigned char)(data >> 8);

    if(addr == 0xA000)  {
        paging_regs[3] = (unsigned char)data;
        remap_page2();
    }
    else if(addr == 0x9FFF) {
        paging_regs[3] = data >> 8;
        remap_page2();
    }
}

static void eeprom_mwrite16(unsigned short addr, unsigned short data)   {
    int top = addr >> 8, bot = addr & 0xFF;

    write_map[top][bot++] = (unsigned char)data;

    if(bot <= 0xFF)
        write_map[top][bot] = (unsigned char)(data >> 8);
    else
        write_map[(unsigned char)(top + 1)][0] = (unsigned char)(data >> 8);

    if(e93c46.enabled)  {
        if(addr == 0x8000)    {
            eeprom93c46_write((unsigned char)data);
        }
        else if(addr == 0x7FFF)   {
            eeprom93c46_write(data >> 8);
        }
        else if(addr >= 0x8008 && addr < 0x8088)    {
            int a = (addr - 0x8008) >> 1;

            if(addr & 0x01) {
                e93c46.data[a] &= 0x00FF;
                e93c46.data[a] |= data << 8;

                if(a < 0x80)    {
                    e93c46.data[a + 1] &= 0xFF00;
                    e93c46.data[a + 1] |= (unsigned char)data;
                }
            }
            else    {
                e93c46.data[a] = data;
            }
        }
    }

    if(addr == 0xFFFC)  {
        eeprom93c46_ctl_write(data);
    }
    else if(addr == 0xFFFB) {
        eeprom93c46_ctl_write(data >> 8);
    }

    if(addr > 0xFFFC)  {
        /* This is bound for a paging register */
        paging_regs[addr - 0xFFFC] = data;
        remap_page[addr - 0xFFFC]();
    }

    if(addr > 0xFFFB)   {
        paging_regs[addr - 0xFFFC + 1] = data >> 8;
        remap_page[addr - 0xFFFC + 1]();
    }
}

void sms_port_write(unsigned short port, unsigned char data)    {
    port &= 0xFF;

    if(port < 0x07 && sms_console == CONSOLE_GG)    {
        gg_regs[port] = data;

        if(port == 0x06)    {
            sn76489_set_output_channels(&psg, data);
        }
    }
    else if(port < 0x40)    {
        if(port & 0x01) {
            /* I/O Control register */
        }
        else    {
            /* Memory Control register */
        }
    }
    else if(port < 0x80)    {
        /* SN76489 PSG */
        sn76489_write(&psg, data);
    }
    else if(port < 0xC0)    {
        if(port & 0x01) {
            /* VDP Control port */
            sms_vdp_ctl_write(data);
        }
        else    {
            /* VDP Data port */
            sms_vdp_data_write(data);
        }
    }
    else    {
    }
}

unsigned char sms_port_read(unsigned short port)    {
    port &= 0xFF;

    if(port < 0x07 && sms_console == CONSOLE_GG)    {
        return gg_regs[port];
    }
    else if(port < 0x40)    {
        return 0xFF;
    }
    else if(port < 0x80)    {
        if(port & 0x01) {
            return sms_vdp_hcnt_read();
        }
        else    {
            return sms_vdp_vcnt_read();
        }
    }
    else if(port < 0xC0)    {
        if(port & 0x01) {
            return sms_vdp_status_read();
        }
        else    {
            return sms_vdp_data_read();
        }
    }
    else    {
        if(port & 0x01) {
            /* I/O port B/misc register */
            return (sms_pad >> 8) & 0xFF;
        }
        else    {
            /* I/O port A/B register */
            return (sms_pad & 0xFF);
        }
    }
}

#ifndef _arch_dreamcast

int sms_write_cartram_to_file(void) {
    FILE *fp;

    if(savefn == NULL)
        return -1;

    if(mapper == SMS_MAPPER_93C46)  {
        fp = fopen(savefn, "wb");

        if(!fp)
            return -1;

        fwrite(e93c46.data, 64, 2, fp);
        fclose(fp);
    }
    else    {
        if(!cartram_enabled)
            return 0;

        fp = fopen(savefn, "wb");
        if(!fp)
            return -1;

        fwrite(cart_ram, 0x8000, 1, fp);
        fclose(fp);
    }

    return 0;
}

int sms_read_cartram_from_file(void)    {
    FILE * fp;

    fp = fopen(savefn, "rb");
    
    if(fp != NULL)  {
        if(mapper == SMS_MAPPER_93C46)  {
            eeprom93c46_init();
            fread(e93c46.data, 64, 2, fp);
        }
        else    {
            fread(cart_ram, 0x8000, 1, fp);
            cartram_enabled = 1;
        }
        fclose(fp);

        return 0;
    }

    return -1;
}

#else

int sms_write_cartram_to_file(void) {
    vmu_pkg_t pkg;
    uint8 *pkg_out;
    uint8 *comp;
    int pkg_size;
    uint32 len;
    FILE *fp;
    char prodcode[7];
    char savename[15];
    int err;

    if(savefn == NULL)  {
        return -1;
    }

    if(!cartram_enabled && mapper != SMS_MAPPER_93C46)  {
        return 0;
    }

    sprintf(prodcode, "%d%d%d%d%d", (cart_rom[0x7FFE] & 0xF0) >> 4,
            (cart_rom[0x7FFD] & 0xF0) >> 4, (cart_rom[0x7FFD] & 0x0F),
            (cart_rom[0x7FFC] & 0xF0) >> 4, (cart_rom[0x7FFC] & 0x0F));

    sprintf(savename, "/vmu/a1/%s", prodcode);

    comp = (uint8 *)malloc(0x10000);
    len = 0x10000;

    if(mapper == SMS_MAPPER_93C46)  {
        err = compress2(comp, &len, (uint8 *)e93c46.data, 128, 9);
    }
    else    {
        err = compress2(comp, &len, cart_ram, 0x8000, 9);
    }

    strcpy(pkg.desc_short, "CrabEmu Save");
    strcpy(pkg.desc_long, prodcode);
    strcpy(pkg.app_id, "CrabEmu");
    pkg.icon_cnt = 1;
    pkg.icon_anim_speed = 0;
    memcpy(pkg.icon_pal, icon_pal, 32);
    pkg.icon_data = icon_img;
    pkg.eyecatch_type = VMUPKG_EC_NONE;
    pkg.data_len = len;
    pkg.data = comp;

    vmu_pkg_build(&pkg, &pkg_out, &pkg_size);

    fp = fopen(savename, "wb");

    if(!fp) {
        free(pkg_out);
        free(comp);
        return -1;
    }

    fwrite(pkg_out, pkg_size, 1, fp);
    fclose(fp);

    free(pkg_out);
    free(comp);

    return 0;
}

int sms_read_cartram_from_file(void)    {
    vmu_pkg_t pkg;
    uint8 *pkg_out;
    int pkg_size;
    char prodcode[7];
    FILE *fp;
    char savename[15];
    uint32 real_size = mapper == SMS_MAPPER_93C46 ? 128 : 0x8000;

    if(savefn == NULL)
        return -1;

    sprintf(prodcode, "%d%d%d%d%d", (cart_rom[0x7FFE] & 0xF0) >> 4,
            (cart_rom[0x7FFD] & 0xF0) >> 4, (cart_rom[0x7FFD] & 0x0F),
            (cart_rom[0x7FFC] & 0xF0) >> 4, (cart_rom[0x7FFC] & 0x0F));
    
    sprintf(savename, "/vmu/a1/%s", prodcode);

    fp = fopen(savename, "rb");

    if(!fp) {
        return -1;
    }

    fseek(fp, 0, SEEK_SET);
    fseek(fp, 0, SEEK_END);
    pkg_size = ftell(fp);
    fseek(fp, 0, SEEK_SET);

    pkg_out = (uint8 *)malloc(pkg_size);
    fread(pkg_out, pkg_size, 1, fp);
    fclose(fp);

    vmu_pkg_parse(pkg_out, &pkg);

    if(mapper == SMS_MAPPER_93C46)  {
        eeprom93c46_init();
        uncompress((uint8 *)e93c46.data, &real_size, pkg.data, pkg.data_len);
    }
    else    {
        uncompress(cart_ram, &real_size, pkg.data, pkg.data_len);
    }

    free(pkg_out);

    cartram_enabled = 1;

    return 0;
}

#endif /* _arch_dreamcast */

int sms_mem_load_rom(const char *fn)    {
    FILE *fp;
    int size, i;
    char *ext;

    memset(cart_ram, 0, 0x8000);
    cartram_enabled = 0;

    fp = fopen(fn, "rb");

    if(!fp) {
#ifdef DEBUG
        fprintf(stderr, "sms_mem_load_rom: Could not open ROM: %s!\n", fn);
#endif
        return -1;
    }

    /* First, determine which type of ROM it is */
    ext = strrchr(fn, '.');

    if(!strcmp(ext, ".gg")) {
        /* GameGear rom */
        sms_console = CONSOLE_GG;
        gui_set_viewport(160, 144);
    }
    else if(!strcmp(ext, ".sc"))    {
        /* SC-3000 rom */
        sms_console = CONSOLE_SG1000;
        gui_set_viewport(256, 192);
    }
    else if(!strcmp(ext, ".sg"))    {
        /* SG-1000 rom */
        sms_console = CONSOLE_SG1000;
        gui_set_viewport(256, 192);
    }
    else    {
        /* Master System rom, hopefully */
        sms_console = CONSOLE_SMS;
        gui_set_viewport(256, 192);
    }

    /* Determine the size of the file */
    fseek(fp, 0, SEEK_END);
    size = ftell(fp);
    fseek(fp, 0, SEEK_SET);

    /* Determine if header/footer info is present */
    if((size % 0x4000) == 512)  {
        /* SMD header present, skip it */
        fseek(fp, 512, SEEK_SET);
        size -= 512;
    }
    else if((size % 0x4000) == 64)  {
        /* Footer present, truncate it */
        size -= 64;
    }

    cart_rom = (unsigned char *)malloc(size);
    fread(cart_rom, size, 1, fp);
    fclose(fp);

    if(savefn)  {
        free(savefn);
    }

    cart_len = size;

    /* Determine which mapper should be used, and set up the paging
       accordingly. First check the mappers database */
    if((mapper = sms_find_mapper(cart_rom, cart_len)) == -1)    {
        mapper = SMS_MAPPER_SEGA;

        if(sms_console != CONSOLE_SG1000)   {
            setup_mapper();

            if(mapper == SMS_MAPPER_SEGA)   {
                /* Check if this game uses the 93c46 serial eeprom, and if so
                   enable it. Most of these should be detected by the
                   sms_find_mapper function, but just in case, detect them
                   here as well. */
                const uint16 prodcodes[5] = {
                    0x3432, 0x2537, 0x2439, 0x3407, 0x2418
                };
                const uint16 popbreaker = 0x2017;
                uint16 prodcode = cart_rom[0x7FFC] | (cart_rom[0x7FFD] << 8) |
                    ((cart_rom[0x7FFE] & 0xF0) << 12);

                for(i = 0; i < 5; ++i)  {
                    if(prodcode == prodcodes[i])    {
                        mapper = SMS_MAPPER_93C46;
                        break;
                    }
                }

                /* While we're at it, check if its Pop Breaker, which apparently
                   doesn't like running on a non-japanese Game Gear */
                if(prodcode == popbreaker)  {
                    gg_regs[0] = 0x80;
                }
            }
        }
    }

    if(sms_console == CONSOLE_SG1000)   {
        for(i = 0; i < 0x08; ++i)   {
            read_map[i + 0xC8] = ram + (i << 8);
            read_map[i + 0xD0] = ram + (i << 8);
            read_map[i + 0xD8] = ram + (i << 8);
            read_map[i + 0xE8] = ram + (i << 8);
            read_map[i + 0xF0] = ram + (i << 8);
            read_map[i + 0xF8] = ram + (i << 8);

            write_map[i + 0xC8] = ram + (i << 8);
            write_map[i + 0xD0] = ram + (i << 8);
            write_map[i + 0xD8] = ram + (i << 8);
            write_map[i + 0xE8] = ram + (i << 8);
            write_map[i + 0xF0] = ram + (i << 8);
            write_map[i + 0xF8] = ram + (i << 8);
        }
    }

    switch(mapper)  {
        case SMS_MAPPER_SEGA:
            sms_z80_set_mread(&normal_mread);
            sms_z80_set_mread16(&normal_mread16);

            if(sms_console != CONSOLE_SG1000)   {
                sms_z80_set_mwrite(&sega_mwrite);
                sms_z80_set_mwrite16(&sega_mwrite16);
            }
            else    {
                sms_z80_set_mwrite(&sg1000_mwrite);
                sms_z80_set_mwrite16(&sg1000_mwrite16);
            }
            
            break;

        case SMS_MAPPER_CODEMASTERS:
            sms_z80_set_mread(&normal_mread);
            sms_z80_set_mwrite(&codemasters_mwrite);
            sms_z80_set_mread16(&normal_mread16);
            sms_z80_set_mwrite16(&codemasters_mwrite16);
            break;

        case SMS_MAPPER_KOREAN:
            sms_z80_set_mread(&normal_mread);
            sms_z80_set_mwrite(&korean_mwrite);
            sms_z80_set_mread16(&normal_mread16);
            sms_z80_set_mwrite16(&korean_mwrite16);
            break;

        case SMS_MAPPER_93C46:
            eeprom93c46_init();
            sms_z80_set_mread(&eeprom_mread);
            sms_z80_set_mwrite(&eeprom_mwrite);
            sms_z80_set_mread16(&eeprom_mread16);
            sms_z80_set_mwrite16(&eeprom_mwrite16);
            break;

        case SMS_MAPPER_CASTLE:
            sms_z80_set_mread(&normal_mread);
            sms_z80_set_mwrite(&sg1000_mwrite);
            sms_z80_set_mread16(&normal_mread16);
            sms_z80_set_mwrite16(&sg1000_mwrite16);
            break;

        case SMS_MAPPER_TEREBI_OEKAKI:
            sms_z80_set_mread(&terebi_mread);
            sms_z80_set_mwrite(&terebi_mwrite);
            sms_z80_set_mread16(&terebi_mread16);
            sms_z80_set_mwrite16(&terebi_mwrite16);
            break;
    }

    reorganize_pages();

    savefn = (char *) malloc(strlen(fn) + 5);
    sprintf(savefn, "%s.ces", fn);

    sms_read_cartram_from_file();

    return 0;
}

void sms_mem_write_context(FILE *fp)    {
    fwrite(paging_regs, 4, 1, fp);
    fwrite(gg_regs, 7, 1, fp);
    fwrite(ram, 0x2000, 1, fp);
    fwrite(cart_ram, 0x8000, 1, fp);
}

void sms_mem_read_context(FILE *fp) {
    fread(paging_regs, 4, 1, fp);
    fread(gg_regs, 7, 1, fp);
    fread(ram, 0x2000, 1, fp);
    fread(cart_ram, 0x8000, 1, fp);
    reorganize_pages();
}

int sms_mem_init(void)  {
    int i;

    cart_rom = NULL;
    rom_page0 = NULL;
    rom_page1 = NULL;
    rom_page2 = NULL;

    remap_page[0] = &remap_page2;
    remap_page[1] = &remap_page0;
    remap_page[2] = &remap_page1;
    remap_page[3] = &remap_page2;

    paging_regs[0] = 0;
    paging_regs[1] = 0;
    paging_regs[2] = 1;
    paging_regs[3] = 0;

    gg_regs[0] = 0xC0;
    gg_regs[1] = 0x7F;
    gg_regs[2] = 0xFF;
    gg_regs[3] = 0x00;
    gg_regs[4] = 0xFF;
    gg_regs[5] = 0x00;
    gg_regs[6] = 0xFF;

    for(i = 0x00; i < 0xC0; ++i)    {
        write_map[i] = dummy_area;
    }

    for(i = 0xC0; i < 0xE0; ++i)    {
        read_map[i] = ram + ((i - 0xC0) << 8);
        write_map[i] = ram + ((i - 0xC0) << 8);
    }

    for(i = 0xE0; i < 0x100; ++i)   {
        read_map[i] = ram + ((i - 0xE0) << 8);
        write_map[i] = ram + ((i - 0xE0) << 8);
    }

    return 0;
}

int sms_mem_shutdown(void)  {
    if(cart_rom != NULL)
        free(cart_rom);

    return 0;
}

void sms_mem_reset(void)    {
    paging_regs[0] = 0;
    paging_regs[1] = 0;
    paging_regs[2] = 1;
    paging_regs[3] = 0;

    if(gg_regs[0] & 0x40)
        gg_regs[0] = 0xC0;
    else
        gg_regs[0] = 0x80;

    gg_regs[1] = 0x7F;
    gg_regs[2] = 0xFF;
    gg_regs[3] = 0x00;
    gg_regs[4] = 0xFF;
    gg_regs[5] = 0x00;
    gg_regs[6] = 0xFF;

    reorganize_pages();
}
